UNIT PCMD1;  {Thresholded derivative and selective sharpening}

INTERFACE

{$R-}   { range checking off}
{$SC+}  { short circuit AND & OR statements in IF's}
USES memtypes, OSIntf, ToolIntf;

CONST
	MaxPixelsPerLine = 2048;
	WhiteC = 0;
	BlackC = 255;
	{returnCodes}
	No_Change			  = 	 0;
	CMDperiod			  = -10000;
	Changed				  = -20000;
	CMDperiod_and_changed = -30000;
	
TYPE
	 PCmdBlock = RECORD
	 				Primary : ptr; {all changes are made to primary}
					Secondary : ptr; {NIL if it couldn't load, or didn't exist}
					OptionKeyDownCall : boolean;
					ShiftKeyDownCall : boolean;
					NLines : integer;
					PixelsPerLine : integer;
					BytesPerRow : integer;
					sPixelsPerLine : integer;
					sBytesPerRow : integer;
					sNLines : integer;
					RoiRect : rect;
					Rectangular : boolean; {T/F}
					mask : ptr;
					ReturnCode : OSErr; {0=okay, <0 = changed,CommandPeriod, or both}
					FgColor : integer;
					iScale:extended;   {extended - # of pixels per unit}
					iunits:string[2];  {2 characters - calibration units}
					MyOSPort : GrafPtr;
	 			 END;
	 PCmdBlockPtr = ^PcmdBlock;
				 
  	UnsignedByte=0..255;
  	LineType=PACKED ARRAY[0..MaxPixelsperLine] OF UnsignedByte;
	SArray=Array[1..25] of integer;
	
				 
{ MAIN entry point}


PROCEDURE  DoPCommand (pData1:PCmdBlockptr);

IMPLEMENTATION

PROCEDURE DoPCommand (pData1:PCmdBlockptr);
VAR
	pData : PCMDBlock;
	
	PROCEDURE dumbBeep;
	BEGIN
		sysbeep(1);
	END;
	
	FUNCTION blankLine : linetype;
	VAR i:integer;
		jl : linetype;
	BEGIN
		for i:=0 to MaxPixelsPerLine-1 do
			jl[i]:=whiteC;
		blankline:=jl;
	END; {blankline}
	
	PROCEDURE GetLine(h,v,count:integer; VAR line:LineType);
	VAR
		offset:LongInt;
		p:ptr;
	BEGIN
		IF pdata.primary=NIL then
			BEGIN
				line:=BlankLine;
				exit(GetLine);
			END;			
		IF (h<0) OR (v<0) OR
		   ((h+count)>pdata.PixelsPerLine)
		   OR (v>=pdata.nlines) THEN
			BEGIN
				line:=BlankLine;
				exit(GetLine);
			END;
		offset:=LongInt(v)*pdata.BytesPerRow+h;
		p:=ptr(ord4(pdata.Primary)+offset);
		BlocKMove(p,@line,count);
	END; {getline}

	PROCEDURE GetBGLine(h,v,count:integer; VAR line:LineType);
	VAR
		offset:LongInt;
		p:ptr;
	BEGIN
		IF pdata.secondary=NIL then
			BEGIN
				line:=BlankLine;
				exit(GetBGLine);
			END;			
		IF (h<0) OR (v<0) OR
		   ((h+count)>pdata.sPixelsPerLine)
		   OR (v>=pdata.snlines) THEN
			BEGIN
				line:=BlankLine;
				exit(GetBGLine);
			END;
		offset:=LongInt(v)*pdata.sBytesPerRow+h;
		p:=ptr(ord4(pdata.secondary)+offset);
		BlocKMove(p,@line,count);
	END; {getBGline}

	PROCEDURE GetMaskLine(h,v,count:integer; VAR line:LineType);
	VAR
		offset:LongInt;
		p:ptr;
	BEGIN
		IF pdata.mask=NIL then
			BEGIN
				line:=BlankLine;
				exit(GetMaskLine);
			END;			
		IF (h<0) OR (v<0) OR
		   ((h+count)>pdata.PixelsPerLine)
		   OR (v>=pdata.nlines) THEN
			BEGIN
				line:=BlankLine;
				exit(GetMaskLine);
			END;
		offset:=LongInt(v)*pdata.BytesPerRow+h;
		p:=ptr(ord4(pdata.mask)+offset);
		BlocKMove(p,@line,count);
	END; {getMaskline}
	
	PROCEDURE PutLine(h,v,count:integer; VAR line:LineType);
	VAR
		offset:LongInt;
		p:ptr;
	BEGIN
		if pdata.primary=NIL then exit(putLine);
		IF (h<0) OR (v<0) OR (v>=pdata.nlines) THEN exit(PutLine);
		IF (h+count)>pdata.PixelsPerLine THEN count:=pdata.PixelsPerLine-h;
		offset:=LongInt(v)*pdata.BytesPerRow+h;
		p:=ptr(ord4(pdata.primary)+offset);
		BlocKMove(@line,p,count);
	END;{PutLine}

	PROCEDURE GetLineUsingMask(h,v,count:integer; var line:linetype;  padcolor:integer);
	VAR {only needed when eroding/dilating & the boundaries are 
	     assumed to be a special color}
		line2 : linetype;
		i: integer;
	BEGIN
		getline (h,v,count,line);
		getline(h,v,count,line2);
		FOR i:=0 to count-1 DO
			if line2[i]<>pdata.fgcolor THEN line[i]:=padcolor;
	END;{GetLineUsingMask}
	
	PROCEDURE PutLineUsingMask(h,v,count:integer; VAR line:LineType);
	VAR
		aLine,MaskLine:LineType;
		i:integer;
	BEGIN
		GetLine(h,v,count,aline);
		GetMaskLine(h,v,count,MaskLine);
		FOR i:=0 TO count-1 DO
		IF MaskLine[i]=pdata.fgcolor THEN aLine[i]:=line[i];
		PutLine(h,v,count,aLine);
	END;

	FUNCTION CommandPeriod:boolean;
	TYPE
		KeyPtrType=^KeyMap;
	VAR
		KeyPtr:KeyPtrType;
		keys:ARRAY[0..3] OF LongInt;
		event : eventrecord;
	BEGIN
	  systemtask;
	  KeyPtr:=KeyPtrType(@keys);
	  GetKeys(KeyPtr^);
	  CommandPeriod:=(BAND(keys[1],$808000))=$808000;
	END;
	
	PROCEDURE SetDialogItem(TheDialog:DialogPtr; item,value:integer);
	VAR
		ItemType:integer;
		ItemBox:rect;
		ItemHdl:handle;
	BEGIN
		GetDItem (TheDialog,item,ItemType,ItemHdl,ItemBox);
		SetCtlValue(ControlHandle(ItemHdl),value)
	END;{SetDialogItem}

	PROCEDURE ShowWatch;
	VAR
		watch:CursHandle;
	BEGIN
		watch := GetCursor(WatchCursor);
		SetCursor(watch^^);
	END;

	PROCEDURE OutlineButton(theDialog: DialogPtr; itemNo, CornerRad: integer);
	{ Draws a border around a button. 16 is the normal
	  cornerRad for small buttons }
	VAR
		itemType: Integer;
		itemBox: Rect;
		itemHdl: Handle;
		tempPort: GrafPtr;
	BEGIN
		GetPort(tempPort);
		SetPort(theDialog);
		GetDItem(theDialog, itemNo, itemType, itemHdl, itemBox);
		PenSize(3, 3);
		InSetRect(itemBox, -4, -4);
		FrameRoundRect(itemBox, cornerRad, cornerRad);
		PenSize(1,1);
		SetPort(tempPort);
	END;{OutlineButton}

VAR
	c,h,i,j : integer; {for scanning entire area}
	irect : rect;
	L1,L2,L3,L4,L5 : linetype; {space for 5 incoming lines}
	Result : linetype;         {result from processing}
	ptr1,ptr2,ptr3,ptr4,ptr5, shuffle : ^linetype;  {pointing at all 5 lines}
	width,height : integer;
	DArray:array[1..4] of longint;
	k,m,swap,s,t : integer;
	mylog : dialogptr;
	
BEGIN
	pdata:=pdata1^;
	
	t:=128;
	j:=6;
	initcursor;
	mylog:=GetNewDialog(2000,nil,pointer(-1));
    OutlineButton(MyLog,ok,16);
    SetDialogItem(Mylog,j,1);
    Repeat {until OK or cancel}
    	ModalDialog(Nil,i);
    	IF (i>=3) AND (i<=10) THEN
                BEGIN
                    SetDialogItem(Mylog,j,0);
                    j:=i;
                    SetDialogItem(MyLog,j,1);
                END;
	Until (i=OK) or (i=cancel);
	DisposDialog(mylog);
	IF (i=cancel) THEN
		BEGIN
			pdata.ReturnCode:=pdata.ReturnCode+CMDperiod;
			pdata1^:=pdata;
			Exit(DoPCommand);
		END;
	case j of
		3:t:=512;
		4:t:=256;
		6:t:=128;
		7:t:=64;
		8:t:=32;
		9:t:=16;
		10:t:=8;
		end; {case}
	
	pdata.ReturnCode:=changed;
	
	irect:=pdata.roirect; {bottom is not inclusive in process area}
	with irect,pdata DO
		BEGIN
			if top<2 then top:=2; {avoid top 2 lines}
			if left<2 then left:=2;
			if bottom>nlines-2 then bottom:=nlines-2;
			if right>PixelsperLine-2 then right:=PixelsPerLine-2;
			width:=right-left;
			height:=bottom-top;
		END;
	
	{set up top 4 lines before entering loop}
	
	ptr1:=@L1;ptr2:=@L2;ptr3:=@L3;ptr4:=@L4;ptr5:=@L5; shuffle:=NIL;
	
			GetLine(irect.left-2,irect.top-2,width+4,ptr1^);
			GetLine(irect.left-2,irect.top-1,width+4,ptr2^);
			GetLine(irect.left-2,irect.top  ,width+4,ptr3^);
			GetLine(irect.left-2,irect.top+1,width+4,ptr4^);
		
	{ top of the loop}
	
	FOR i:=irect.top to irect.bottom-1 DO
		BEGIN
			GetLine(irect.left-2,i+2,width+4,ptr5^);
			
			for j:=2 to width+1 DO
				BEGIN
			
					DArray[ 1]:=abs(ptr1^[j-2]+ptr1^[j-1]+ptr2^[j-2]+ptr2^[j-1]
							-ptr4^[j+1]-ptr4^[j+2]-ptr5^[j+1]-ptr5^[j+2]);
					DArray[ 2]:=abs(ptr3^[j-2]+ptr3^[j-1]+ptr2^[j-2]+ptr4^[j-2]
							-ptr3^[j+1]-ptr3^[j+2]-ptr2^[j+2]-ptr4^[j+2]);
					DArray[ 3]:=abs(ptr5^[j-2]+ptr5^[j-1]+ptr4^[j-2]+ptr4^[j-1]
							-ptr4^[j+1]-ptr4^[j+2]-ptr5^[j+1]-ptr5^[j+2]);
					DArray[ 4]:=abs(ptr1^[j  ]+ptr2^[j  ]+ptr1^[j-1]+ptr1^[j+1]
							-ptr5^[j  ]-ptr4^[j  ]-ptr5^[j-1]-ptr5^[j+1]);
					c:=DArray[1];s:=1;
					For m:=2 to 4 DO
						IF DArray[j]>c then
							BEGIN
								s:=m;
								c:=DArray[m];
							END;
					IF c>t THEN
						BEGIN
							k:=3*ptr3^[j];
							Case s of
								1:k:=k-ptr2^[j-1]-ptr4^[j+1];
								2:k:=k-ptr3^[j-1]-ptr3^[j+1];
								3:k:=k-ptr2^[j+1]-ptr4^[j-1];
								4:k:=k-ptr2^[ j ]-ptr4^[ j ];
								end;
						END Else k:=ptr3^[j];
									
					if k>255 then Result[j-2]:=255
								ELSE if k<0 THEN Result[j-2]:=0
									ELSE Result[j-2]:=trunc(k);
				END; {for j}
				
			if pdata.rectangular THEN
				PutLine(irect.left,i,width,result)
			ELSE
				PutLineUsingMask(irect.left,i,width,result);
			
			shuffle:=ptr1;
			ptr1:=ptr2;
			ptr2:=ptr3;
			ptr3:=ptr4;
			ptr4:=ptr5;
			ptr5:=shuffle;
			
			IF commandperiod then
				BEGIN
					sysbeep(1);
					pdata.ReturnCode:=pdata.ReturnCode+CMDperiod;
					pdata1^:=pdata;
					Exit(DoPCommand);
				END;
		END; {for i}
		
		
		pData1^:=pdata;
	
END; {DoPCommand}



END. {UNIT}