UNIT PCMD1;  {3x3 truncated median filter}

INTERFACE

{$R-}   { range checking off}
{$SC+}  { short circuit AND & OR statements in IF's}
USES memtypes, OSIntf, ToolIntf;

CONST
	MaxPixelsPerLine = 2048;
	WhiteC = 0;
	BlackC = 255;
	{returnCodes}
	No_Change			  = 	 0;
	CMDperiod			  = -10000;
	Changed				  = -20000;
	CMDperiod_and_changed = -30000;
	
TYPE
	 PCmdBlock = RECORD
	 				Primary : ptr; {all changes are made to primary}
					Secondary : ptr; {NIL if it couldn't load, or didn't exist}
					OptionKeyDownCall : boolean;
					ShiftKeyDownCall : boolean;
					NLines : integer;
					PixelsPerLine : integer;
					BytesPerRow : integer;
					sPixelsPerLine : integer;
					sBytesPerRow : integer;
					sNLines : integer;
					RoiRect : rect;
					Rectangular : boolean; {T/F}
					mask : ptr;
					ReturnCode : OSErr; {0=okay, <0 = changed,CommandPeriod, or both}
					FgColor : integer;
					iScale:extended;   {extended - # of pixels per unit}
    				iunits:string[2];  {2 characters - hold the current 
										calibration units}
					MyOSPort : GrafPtr;
	 			 END;
	 PCmdBlockPtr = ^PcmdBlock;
				 
  	UnsignedByte=0..255;
  	LineType=PACKED ARRAY[0..MaxPixelsperLine] OF UnsignedByte;
	SArray=Array[1..25] of integer;
	
				 
{ MAIN entry point}


PROCEDURE  DoPCommand (pData1:PCmdBlockptr);

IMPLEMENTATION

PROCEDURE DoPCommand (pData1:PCmdBlockptr);
VAR
	pData : PCMDBlock;
	
	PROCEDURE dumbBeep;
	BEGIN
		sysbeep(1);
	END;
	
	FUNCTION blankLine : linetype;
	VAR i:integer;
		jl : linetype;
	BEGIN
		for i:=0 to MaxPixelsPerLine-1 do
			jl[i]:=whiteC;
		blankline:=jl;
	END; {blankline}
	
	PROCEDURE GetLine(h,v,count:integer; VAR line:LineType);
	VAR
		offset:LongInt;
		p:ptr;
	BEGIN
		IF pdata.primary=NIL then
			BEGIN
				line:=BlankLine;
				exit(GetLine);
			END;			
		IF (h<0) OR (v<0) OR
		   ((h+count)>pdata.PixelsPerLine)
		   OR (v>=pdata.nlines) THEN
			BEGIN
				line:=BlankLine;
				exit(GetLine);
			END;
		offset:=LongInt(v)*pdata.BytesPerRow+h;
		p:=ptr(ord4(pdata.Primary)+offset);
		BlocKMove(p,@line,count);
	END; {getline}

	PROCEDURE GetBGLine(h,v,count:integer; VAR line:LineType);
	VAR
		offset:LongInt;
		p:ptr;
	BEGIN
		IF pdata.secondary=NIL then
			BEGIN
				line:=BlankLine;
				exit(GetBGLine);
			END;			
		IF (h<0) OR (v<0) OR
		   ((h+count)>pdata.sPixelsPerLine)
		   OR (v>=pdata.snlines) THEN
			BEGIN
				line:=BlankLine;
				exit(GetBGLine);
			END;
		offset:=LongInt(v)*pdata.sBytesPerRow+h;
		p:=ptr(ord4(pdata.secondary)+offset);
		BlocKMove(p,@line,count);
	END; {getBGline}

	PROCEDURE GetMaskLine(h,v,count:integer; VAR line:LineType);
	VAR
		offset:LongInt;
		p:ptr;
	BEGIN
		IF pdata.mask=NIL then
			BEGIN
				line:=BlankLine;
				exit(GetMaskLine);
			END;			
		IF (h<0) OR (v<0) OR
		   ((h+count)>pdata.PixelsPerLine)
		   OR (v>=pdata.nlines) THEN
			BEGIN
				line:=BlankLine;
				exit(GetMaskLine);
			END;
		offset:=LongInt(v)*pdata.BytesPerRow+h;
		p:=ptr(ord4(pdata.mask)+offset);
		BlocKMove(p,@line,count);
	END; {getMaskline}
	
	PROCEDURE PutLine(h,v,count:integer; VAR line:LineType);
	VAR
		offset:LongInt;
		p:ptr;
	BEGIN
		if pdata.primary=NIL then exit(putLine);
		IF (h<0) OR (v<0) OR (v>=pdata.nlines) THEN exit(PutLine);
		IF (h+count)>pdata.PixelsPerLine THEN count:=pdata.PixelsPerLine-h;
		offset:=LongInt(v)*pdata.BytesPerRow+h;
		p:=ptr(ord4(pdata.primary)+offset);
		BlocKMove(@line,p,count);
	END;{PutLine}

	PROCEDURE GetLineUsingMask(h,v,count:integer; var line:linetype;  padcolor:integer);
	VAR {only needed when eroding/dilating & the boundaries are 
	     assumed to be a special color}
		line2 : linetype;
		i: integer;
	BEGIN
		getline (h,v,count,line);
		getline(h,v,count,line2);
		FOR i:=0 to count-1 DO
			if line2[i]<>pdata.fgcolor THEN line[i]:=padcolor;
	END;{GetLineUsingMask}
	
	PROCEDURE PutLineUsingMask(h,v,count:integer; VAR line:LineType);
	VAR
		aLine,MaskLine:LineType;
		i:integer;
	BEGIN
		GetLine(h,v,count,aline);
		GetMaskLine(h,v,count,MaskLine);
		FOR i:=0 TO count-1 DO
		IF MaskLine[i]=pdata.fgcolor THEN aLine[i]:=line[i];
		PutLine(h,v,count,aLine);
	END;

	FUNCTION CommandPeriod:boolean;
	TYPE
		KeyPtrType=^KeyMap;
	VAR
		KeyPtr:KeyPtrType;
		keys:ARRAY[0..3] OF LongInt;
		event : eventrecord;
	BEGIN
	  systemtask;
	  KeyPtr:=KeyPtrType(@keys);
	  GetKeys(KeyPtr^);
	  CommandPeriod:=(BAND(keys[1],$808000))=$808000;
	END;

	PROCEDURE QuickSort(VAR X:SArray; N:Integer);	
		PROCEDURE QSort(VAR X:Sarray; M,N:integer);
		VAR
			i,j,hold:Integer;			
			PROCEDURE Partit(VAR A:SArray; VAR i,j:integer; left,right:integer);
			VAR
				pivot:integer;				
				PROCEDURE Swap(Var P,Q:integer);
				VAR
					Hold:integer;
				BEGIN
					Hold:=P;
					P:=Q;
					Q:=Hold;
				END; {Swap}
			BEGIN {Partit}
				Pivot:=A[(Left+Right) Div 2];
				i:=Left;
				j:=Right;
				While i<=j DO
					BEGIN
						While A[i]<Pivot DO
							i:=i+1;
						While Pivot<A[j] DO
							j:=j-1;
						IF i<=j THEN
							BEGIN
								{Swap(A[i],A[j]);}
								Hold:=A[i];
								A[i]:=A[j];
								A[j]:=Hold;
								i:=i+1;
								j:=j-1;
							END;
					END;{while}
			END;{Partit}
		BEGIN {QSort}
			IF M<N THEN
				BEGIN
					Partit(X,I,J,M,N); {divide in two}
					QSort(X,M,J); {sort left part}
					QSort(X,i,N); {Sort right part}
				END;
		END; {QSort}
	BEGIN{QuickSort}
		QSort(X,1,N);
	END;{QuickSort}

	PROCEDURE ShellSort(VAR A:SArray; N:Integer);
	VAR
		Done:Boolean;
		Jump,i,j,Hold:integer;
	BEGIN
		Jump:=N;
		While Jump>1 DO
			BEGIN
				Jump:=Jump DIV 2;
				Repeat
					Done:=True;
					For j:=1 to n-Jump DO
						BEGIN
							i:=j+Jump;
							IF A[j]>A[i] THEN
								BEGIN
									Hold:=A[j];
									A[j]:=A[i];
									A[i]:=Hold;
									Done:=False;
								END;
						END;
				Until Done;
			END;
	END;

VAR
	i,j : integer; {for scanning entire area}
	irect : rect;
	L1,L2,L3 : linetype; {space for 3 incoming lines}
	Result : linetype;         {result from processing}
	ptr1,ptr2,ptr3, shuffle : ^linetype;  {pointing at all 5 lines}
	width,height : integer;
	sortarray : SArray {array[1..9] of unsignedbyte};
	k,m,swap : integer; {for bubblesort}
	Mean:LongInt;
	
BEGIN {truncated median}
	pdata:=pdata1^;
	
	pdata.ReturnCode:=changed; {if the image is going to change}
	
	irect:=pdata.roirect; {bottom is not inclusive in process area}
	with irect,pdata DO
		BEGIN
			if top<1 then top:=1; {avoid top line}
			if left<1 then left:=1;
			if bottom>nlines-1 then bottom:=nlines-1;
			if right>PixelsperLine-1 then right:=PixelsPerLine-1;
			width:=right-left;
			height:=bottom-top;
		END;
	
	{set up top 2 lines before entering loop}
	
	ptr1:=@L1;ptr2:=@L2;ptr3:=@L3; shuffle:=NIL;
	
			GetLine(irect.left-1,irect.top-1,width+2,ptr1^);
			GetLine(irect.left-1,irect.top  ,width+2,ptr2^);
		
	{ top of the loop}
	
	FOR i:=irect.top to irect.bottom-1 DO
		BEGIN
			GetLine(irect.left-1,i+1,width+2,ptr3^);
			
			for j:=1 to width DO
				BEGIN { *** process each point, w/ Result[] ***}
					SortArray[ 1]:=ptr1^[j-1];
					SortArray[ 2]:=ptr1^[j  ];
					SortArray[ 3]:=ptr1^[j+1];
					SortArray[ 4]:=ptr2^[j-1];
					SortArray[ 5]:=ptr2^[j  ];
					SortArray[ 6]:=ptr2^[j+1];
					SortArray[ 7]:=ptr3^[j-1];
					SortArray[ 8]:=ptr3^[j  ];
					SortArray[ 9]:=ptr3^[j+1];
				
					Mean:=0;
					For k:=1 to 9 DO Mean:=Mean+SortArray[k];
					Mean:=Mean DIV 9;
					
					ShellSort(SortArray,9);
					
					k:=5;
					{discard the two values farthest from the mean}
					IF ABS(SortArray[1]-Mean)<ABS(Mean-SortArray[9])
						THEN BEGIN {discard 9}
							IF ABS(SortArray[1]-Mean)<ABS(Mean-SortArray[8])
								THEN k:=4; {and 8}
						END ELSE BEGIN {discard 1}
							IF ABS(Mean-SortArray[9])<ABS(SortArray[2]-Mean)
								THEN k:=6; {and 2}
						END; {other combinations discard one from each end of
							the distribution and do not change the k=5 rule}

					Result[j-1]:=SortArray[k]; {median or truncated median}
					
					{reference E. R. Davies (1984) "The median filter: an appraisal
					and a new truncated version" Proc. 7th Int. Conf. on Patt. Recog.
					Montreal, p. 590-592 and E. R. Davies (1988) "On the noise
					suppression and image enhancement characteristics of the median,
					truncated median and mode filters" Patt. Recog. Lett. 7:87-97}
					
				END; {for j}
				
			if pdata.rectangular THEN
				PutLine(irect.left,i,width,result)
			ELSE
				PutLineUsingMask(irect.left,i,width,result);
			
			shuffle:=ptr1;
			ptr1:=ptr2;
			ptr2:=ptr3;
			ptr3:=shuffle;
			
			IF commandperiod then
				BEGIN
					sysbeep(1);
					pdata.ReturnCode:=pdata.ReturnCode+CMDperiod;
					pdata1^:=pdata;
					Exit(DoPCommand);
				END;
		END; {for i}
		
		
		pData1^:=pdata;
	
END; {DoPCommand}



END. {UNIT}